//
//  Voice chat audio recording/streaming routines. 
//

#include <assert.h>
#include <nitro.h>
#include <dwc.h>

#include "main.h"
#include "init.h"
#include "dbs.h"
#include "screen.h"
#include "utility.h"

#include "sound.h"
#include "demo_util.h"

#pragma profile off

//
//  Declare audio parameters 
//
#define SAMPLING_RATE       VCT_AUDIO_FRAME_RATE    // = 8000Hz
#define SAMPLING_TIME       VCT_AUDIO_FRAME_LENGTH  // = 68ms
#define SAMPLING_BYTE       2                       // = 16bit

#define WAVE_SAMPLE ((int) (SAMPLING_RATE * SAMPLING_TIME * SAMPLING_BYTE) / 1000)

#define MAX_CHANNELS 3

static u8 sPlayBuffer[WAVE_SAMPLE * 2 * MAX_CHANNELS] ATTRIBUTE_ALIGN(32); // for Playback
static u8  sRecBuffer[WAVE_SAMPLE * 2]                ATTRIBUTE_ALIGN(32); // for Recording

//  Turn on this macro if you want to use software mixer. 
//
#define USE_SOFT_MIXER

//
//  for Debugging beep 
//
#define PSG_BUFFER_LENGTH 1088
static u8 psgData[PSG_BUFFER_LENGTH];
static BOOL sPSG = FALSE;

//
//  Status of each channles for displaying menu. 
//
extern BOOL gChannelStatus[3];

//
//  for NitroSystem 
//
static struct NNSSndStrm sSndStream;
static NNSSndArc sArc;
static NNSSndHandle seHandle;
extern u32 sound_data_sdat[];

extern GameSequence gameSeqList[GAME_MODE_NUM];
extern GameControl stGameCnt;
extern KeyControl stKeyCnt;

/*---------------------------------------------------------------------------*
   Create debugging beep wave. 
 *---------------------------------------------------------------------------*/
static void makePSG(void* buffer, u32 length)
{
	u32 i;
    s16 sample = 10000;

    s16* buf = (s16*)buffer;

    length /= 2;
    for (i = 0; i < length; ++i) {
        buf[i] = sample;
        if (i % 8 != 7) {
            continue;
        }
        sample = (s16)-sample;
    }
}

/*---------------------------------------------------------------------------*
   Streaming callback function (called NNS_SndStrm) 
 *---------------------------------------------------------------------------*/
static void SndCallback(NNSSndStrmCallbackStatus sts,
                        int nChannels,
                        void* buffer[],
                        u32 length,
                        NNSSndStrmFormat format,
                        void* arg)
{
#pragma unused(format)

	OSTick      start;
    const void *micAddr;
    u32         offset;
    u8*         micSrc;
    u32         ch;
    static int  sCurBuffer = 0;

	
    micSrc = (u8*)arg;

    if (sts == NNS_SND_STRM_CALLBACK_SETUP) {
        for (ch = 0; ch < nChannels; ++ch) {
            MI_CpuClear8(buffer[ch], length);
        }
        return;
    }

    start = OS_GetTick();

    //
    //  Check point of recording buffer with MIC_GetLastSamplingAddress 
    //
    micAddr = MIC_GetLastSamplingAddress();
    offset  = (u32)((u8*)micAddr - micSrc);

    if ( offset < length ) {
        micSrc = micSrc + length;
        if (sCurBuffer == 0) { OS_TPrintf("mic buffer is same\n"); }
        sCurBuffer = 0;
    }
    else {
        if (sCurBuffer == 1) { OS_TPrintf("mic buffer is same\n"); }
        sCurBuffer = 1;
    }

    //  Playback debugging beep 
    //
    if (sPSG) {
        micSrc = psgData;
    }

    //
    //
    // You have to call VCT_Send/ReceiveAudio anytime. These functions check
    // VAD and SSP status and send audio packet only when necessary.
    //
    // The length must be 1088 bytes (= 68ms 8KHz/16bit audio). No other value
    // is accepted.
    //
    VCT_SendAudio(micSrc, length);

    //
    // If you don't use software mixer, you need to call VCT_ReceiveAudio
    // with respect each audio channel.
    //
#ifndef  USE_SOFT_MIXER
    for (ch = 0; ch < nChannels; ++ch) {
        // Return value is TRUE, the client of this channel is talking.
        gChannelStatus[ch] = VCT_ReceiveAudio(buffer[ch], length, ch, NULL);
    }
#else
    //
    // If you use software mixer, call VCT_ReceiveAudio only once.
    VCT_ReceiveMixedAudio(buffer[0], length);
    //
    // In this sample, copy audio data to other channels (for crank up).
    //
    if (nChannels > 1) {
        for (ch = 1; ch < nChannels; ++ch) {
            MI_CpuCopyFast(buffer[0], buffer[ch], length);
        }
    }
#endif

    // for display ticks of this process.
    //
    AddAudioTick(OS_GetTick() - start);

    // Clear and flush CPU cache.
    //
    for (ch = 0; ch < nChannels; ++ch) {
        DC_FlushRange(buffer[ch], length);
    }
    DC_InvalidateRange(micSrc, length);
    return;
}

/*---------------------------------------------------------------------------*
  Initialize recording/streaming (only initial)
 *---------------------------------------------------------------------------*/
static void InitFirst()
{
    OS_TPrintf("Init sound system\n");
    // Initialize MIC.
    //
    MIC_Init();
    PM_Init();
    (void)PM_SetAmp(PM_AMP_ON);
    PM_SetAmpGain(PM_AMPGAIN_40);

    // Wait for ready to mic.
    OS_Sleep(500);

    // Initialize NitroSystem sound system.
    //
    NNS_SndInit();
    NNS_SndStrmInit(&sSndStream);

    //  Initalize NitroSystem sound archive system. 
    //
    NNS_SndArcInitOnMemory(&sArc, sound_data_sdat);
    NNS_SndArcPlayerSetup(NULL);
    NNS_SndHandleInit(&seHandle);

    //  Create debugging beep wave. 
    //
    makePSG(psgData, PSG_BUFFER_LENGTH);
    
    OS_TPrintf("Init sound system done.\n");
}

/*---------------------------------------------------------------------------*
   Initialize recording/stream 
 *---------------------------------------------------------------------------*/
static void InitSound(NNSSndStrmCallback func, void *data,
                      int sampling_time, u32 channels)
{
    static BOOL noInit = TRUE;
  
    u8 cArray[3] = {7, 8, 9};
    MICAutoParam micParam;
    BOOL ret;
    NNSSndStrmFormat nns_format;
    u32 length;

    if (noInit) {
        InitFirst();
        noInit = FALSE;
    }
 
    nns_format = NNS_SND_STRM_FORMAT_PCM16;
    length = (u32)(SAMPLING_RATE * sampling_time * SAMPLING_BYTE) / 1000;
    
    //  Initialize MIC parameters. 
    //
    micParam.type   = MIC_SAMPLING_TYPE_SIGNED_12BIT;
    micParam.buffer = sRecBuffer;
    micParam.size   = length * 2;

    // You MUST set mic sampling frequence with NNS_SND_STRM_TIMER_CLOCK.
    // If you use MC_SAMPLING_RATE_*, recording and playback is out of sync.
    //
    //micParam.rate = MIC_SAMPLING_RATE_8K;
    micParam.rate = (u32)((NNS_SND_STRM_TIMER_CLOCK / SAMPLING_RATE) * 64);
    
    micParam.loop_enable = TRUE;
    micParam.full_callback = NULL;
    micParam.full_arg = NULL;
    MIC_StartAutoSampling(&micParam);
        
    //  Initalize sound streaming. 
    //
    //
    NNS_SndStrmAllocChannel(&sSndStream, (int)channels, cArray);
    NNS_SndStrmSetVolume(&sSndStream, 0);

    
    ret = NNS_SndStrmSetup(&sSndStream,
                           nns_format,
                           sPlayBuffer,
                           length * 2 * channels,
                           NNS_SND_STRM_TIMER_CLOCK / SAMPLING_RATE,
                           2,
                           func,
                           data);
    SDK_ASSERT(ret);

    return;
}

/*---------------------------------------------------------------------------*
   Finalize recording/streaming. 
 *---------------------------------------------------------------------------*/
void TerminateSound()
{
    NNS_SndStrmFreeChannel(&sSndStream);
}

/*---------------------------------------------------------------------------*
   Start streaming. 
 *---------------------------------------------------------------------------*/
void StartSoundLoop()
{
    NNS_SndStrmStart(&sSndStream);
}

/*---------------------------------------------------------------------------*
   Stop streaming. 
 *---------------------------------------------------------------------------*/
void StopSoundLoop()
{
    (void)MIC_StopAutoSampling();
    NNS_SndStrmStop(&sSndStream);
}

/*---------------------------------------------------------------------------*
   Playback ringer tone. 
 *---------------------------------------------------------------------------*/
void PlaySE(int id)
{
    NNS_SndArcPlayerStartSeqArc( &seHandle, 0, id);
}

void StopSE()
{
    NNS_SndPlayerStopSeq(&seHandle, 0);
}

/*---------------------------------------------------------------------------*
   Toggle play to debugging beep. 
 *---------------------------------------------------------------------------*/
void SendBeep(BOOL flag)
{
    sPSG = flag;
}

/*---------------------------------------------------------------------------*
   Terminate sound system. 
 *---------------------------------------------------------------------------*/
void SetupSound(u32 channles)
{
    InitSound(SndCallback, sRecBuffer, VCT_AUDIO_FRAME_LENGTH, channles);
}
