//
//  libVCT transceiver mode sample 
//

#include <nitro.h>
#include <ninet/ip.h>
#include <dwc.h>

#include "main.h"
#include "init.h"
#include "dbs.h"
#include "utility.h"
#include "screen.h"

#include "transceiver.h"
#include "sound.h"
#include "vct.h"

#include "vct_menu.h"
#include "demo_util.h"
#include "demo_menu.h"

#ifdef SDK_DEBUG
#include "debug_menu.h"
#endif

//
//  Global variables. 
//
extern GameSequence gameSeqList[GAME_MODE_NUM];
extern GameControl stGameCnt;
extern KeyControl stKeyCnt;

//
//  Static variables. 
//
static VCTSession sSession[1];
static VCTSession *sCurrentSession = NULL;

extern KeyControl stKeyCnt;
extern GameControl stGameCnt;

static Menu *sMenu = NULL;

//
//  Function prototypes. 
//
static void         ClearSession(VCTSession *session);
static Menu*        InitMenu();
static void         DestroyMenu();

/////////////////////////////////////////////////////////////////////////////////////
//
//  Statistics of function cost 
//
#ifdef SDK_PROFILE_FUNCTIONCOST

#include "cost.h"

#define MYCOSTBUFFER_SIZE 0x30000
#define MYSTATBUFFER_SIZE 0x3000

static u32 myCostBuffer[MYCOSTBUFFER_SIZE / sizeof(u32)];
static u32 myStatBuffer[MYSTATBUFFER_SIZE / sizeof(u32)];

#endif

/////////////////////////////////////////////////////////////////////////////////////
//
//  Callback function when L button down.
//
static void LButtonDown()
{
    VCTSession *session;
	int ret;

    ret = VCT_Contact(&session);
    switch (ret) {
    case VCT_ERROR_TRANSCEIVER_BUSY:
        //  If anyone is already talking, the return value is VCT_ERROR_TRANSCEIVER_BUSY. 
        //
        OS_TPrintf("Transceiver is busy\n");
        PlaySE(SE_BUSY);
        break;

    case VCT_ERROR_SEND_FAIL:
        //  If you failed to send request, you need to delete session. 
        //
        VCT_DeleteSession(session);
        break;

    case VCT_ERROR_NONE:
        //  Save current session. 
        sCurrentSession = session;
        break;

    default:
        break;
    }
    
}

/////////////////////////////////////////////////////////////////////////////////////
//
//  Callback function when L button up. 
//
static void LButtonUp()
{
    int ret;

    if (sCurrentSession) {
        //
        // Stop audio streaming before send RELEASE request, because 
        // when the response of the server is delayed, audio streaming
        // waste of network bandwidth.
        //
        VCT_StopStreaming(sCurrentSession);
        ret = VCT_Release(sCurrentSession);
        if (ret != VCT_ERROR_NONE) {
            OS_TPrintf("Can't send Release request [%d]\n", ret);
        }
    }
}

/////////////////////////////////////////////////////////////////////////////////////
//
//  Event callback of libVCT.
//
static void VoiceChatEventCallback(u8 aid, VCTEvent event, VCTSession *session, void *data)
{
#pragma unused(aid, data)
    
    switch (event) {
    //
    //  Event of transceiver server.
    //
    case VCT_EVENT_CONTACT:
        //
        //  Response of CONTACT is already processed in libVCT.
        //
        OS_TPrintf("Contact from aid:%d\n", session->aid);
        sCurrentSession = session;
        VCT_StartStreaming(session);
        BroadCastStatus(DEMO_STATUS_TALKING);
        PlaySE(SE_NOISE);
        break;

    case VCT_EVENT_RESPONDBYE:
        OS_TPrintf("Release from aid:%d\n", session->aid);
        PlaySE(SE_NOISE);
        VCT_StopStreaming(session);
        ClearSession(session);
        break;

    //
    //  Event of transceiver server and client.
    //
    case VCT_EVENT_DISCONNECTED:
        VCT_StopStreaming(session);
        OS_TPrintf("Disconnected\n");
        ClearSession(session);
        PlaySE(SE_NOISE);
        break;

    //
    //  Event of transceiver client.
    //
    case VCT_EVENT_BUSY:
        OS_TPrintf("Return busy here from aid:%d\n", session->aid);
        ClearSession(session);
        PlaySE(SE_BUSY);
        break;
        
    case VCT_EVENT_CONNECTED:
        OS_TPrintf("Connected by aid:%d (talking %d)\n", session->aid, session->talking);
        PlaySE(SE_BEEP);
        VCT_StartStreaming(session);
        BroadCastStatus(DEMO_STATUS_TALKING);
        break;

    case VCT_EVENT_NOTIFY_FREE:
        OS_TPrintf("Notify Free\n");
        PlaySE(SE_NOISE);
        VCT_StopStreaming(session);
        ClearSession(session);
        break;
        
    case VCT_EVENT_NOTIFY_BUSY:
        OS_TPrintf("Notify Busy\n");
        PlaySE(SE_NOISE);
        VCT_StartStreaming(session);
        BroadCastStatus(DEMO_STATUS_TALKING);
        sCurrentSession = session;
        break;

    case VCT_EVENT_TIMEOUT:
        if (session == sCurrentSession) {
            OS_TPrintf("Session timeout\n");
            VCT_StopStreaming(session);
            ClearSession(session);
        }
        break;

    case VCT_EVENT_ABORT:
        OS_TPrintf("Session abrot\n");
        VCT_StopStreaming(session);
        ClearSession(session);
        break;

    default:
        OS_TPrintf(" not handler (%s)\n", EventName(event));
        break;
    }
}

/////////////////////////////////////////////////////////////////////////////////////
//
//  Clear session
//
static void ClearSession(VCTSession *session)
{
    if (session == sCurrentSession) {
        VCT_StopStreaming(session);
        VCT_DeleteSession(session);
        sCurrentSession = NULL;
    }
    else {
        SDK_WARNING(FALSE, "What is this session??\n");
        VCT_DeleteSession(session);
    }
    sMenu->ChangeMenu(MENU_AUDIO_INFO);
    BroadCastStatus(DEMO_STATUS_TRANSCEIVER);
}

/////////////////////////////////////////////////////////////////////////////////////
//
//  Transceiver mode main routine.
//
GameMode TransceiverMain()
{
    GameSequence *gameSeq = &gameSeqList[GAME_MODE_TRANSCEIVER];
    GameMode returnSeq = GAME_MODE_NUM;
    int curIdx = 0;

    InitMenu();
    DWC_SetReportLevel(0);

#ifdef SDK_PROFILE_FUNCTIONCOST
    OS_InitFunctionCost(&myCostBuffer, MYCOSTBUFFER_SIZE);
    OS_InitStatistics(&myStatBuffer, MYSTATBUFFER_SIZE);
#endif

    BroadCastStatus(DEMO_STATUS_TRANSCEIVER);
    //
    //  Initialize libVCT
    //
    VCTConfig config;

    config.mode         = VCT_MODE_TRANSCEIVER;
    config.session      = sSession;
    config.numSession   = 1;
    config.aid          = DWC_GetMyAID();
    config.callback     = VoiceChatEventCallback;
    config.userData     = NULL;
    config.audioBuffer     = gAudioBuffer;
    config.audioBufferSize = VCT_AUDIO_BUFFER_SIZE * VCT_DEFAULT_AUDIO_BUFFER_COUNT;

    SetupSound(1);

    if (!VCT_Init(&config)) {
        OS_TPrintf("ERROR: Can't initialize VoiceChat!!!\n");
        return GAME_MODE_CONNECTED;
    }

    //  Change transceiver timeout on SDK_DEBUG build.
    //
#ifdef SDK_DEBUG
    VCT_SetTransceiverTimeout(15);
#endif

    //  Re-set connection close callback (re-set returnSeq).
    DWC_SetConnectionClosedCallback(ConnectionClosedCallback, &returnSeq);
    
    sCurrentSession = NULL;

    DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
    
    //  Start sound system.
    StartSoundLoop();

    while (1) {
        dbs_Print( 0, 0, "s:%d", DWC_GetMatchingState() );
        dbs_Print( 7, 0, "n:%d", DWC_GetNumConnectionHost() );
        dbs_Print(14, 0, "a:%d", DWC_GetMyAID());
        dbs_Print( 0, 1, "w:%d", DWC_GetLinkLevel() );
        dbs_Print( 10,1, "p:%d", stGameCnt.userData.profileID );
        dbs_Print( 2, 2, "== Transceiver mode demo ==");

        ReadKeyData();
        DWC_ProcessFriendsMatch();
        
        if (DWC_GetLastError(NULL)){
            //  match make failed...
            ShutdownInet();
            DWC_ClearError();
            returnSeq = GAME_MODE_MAIN;
            break;
        }
        
        //  Call VCT_Main on each picture frame.
        VCT_Main();

        if (returnSeq != GAME_MODE_NUM) break;

        //  Change menu screen with the SELECT button
        if (stKeyCnt.trg & PAD_BUTTON_SELECT) {
            sMenu->GoNext();
        }
        if (stKeyCnt.trg & PAD_BUTTON_L) {
            LButtonDown();
        }
        else if (stKeyCnt.release & PAD_BUTTON_L) {
            LButtonUp();
        }
        
        int key = sMenu->ShowMenu();
        
        //
        //  The 'key' has some data when ShowMenu() did not process.
        //

        if (key & PAD_BUTTON_A) {
            switch (curIdx){
            case 0:  // Client
            case 1:  // Client
            case 2:  // Client
            case 3:  // Client
            case 4:  // Client
            case 5:  // Client
            case 6:  // Client
            case 7:  // Client
		        VCT_SetTransceiverMode(FALSE);
		        VCT_SetTransceiverServer((u8)curIdx);
                break;
            case 8:  // Server
		        u8* aidList;
		        int n = DWC_GetAIDList(&aidList);
		        VCT_SetTransceiverMode(TRUE);
		        VCT_SetTransceiverClients(aidList, n);
                break;
            case 9:  // Return
                returnSeq = GAME_MODE_CONNECTED;
                break;
            }
        }
        else if (key & PAD_KEY_UP) {
            curIdx--;
            if (curIdx < 0) curIdx = gameSeq->numMenu-1;
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (key & PAD_KEY_DOWN) {
            curIdx++;
            if (curIdx >= gameSeq->numMenu) curIdx = 0;
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_BUTTON_DEBUG) {
            //  Trun off backlight for avoid NITRO-Emulator noise.
            (void)PM_SetBackLight( PM_LCD_ALL, PM_BACKLIGHT_OFF );
        }
        
        //  Playback debugging beep wave.
        SendBeep(stKeyCnt.cont & PAD_BUTTON_R);
        
        ShowTicks(sCurrentSession ? StateName(sCurrentSession->state) : "No session");

        //  NitroSystem sound main.
        NNS_SndMain();

        //  Update screen.
        dbs_DemoUpdate();

        //  Wait for V-blank interrupts.
        OS_WaitIrq(TRUE, OS_IE_V_BLANK);
        dbs_DemoLoad();
        
        //  Check stack overflow.
        OS_CheckStack(OS_GetCurrentThread());

#ifdef SDK_PROFILE_FUNCTIONCOST
        __vct_cost_count();
        OS_CalcStatistics(&myStatBuffer);
#endif
    }
    //  Delete all sessions.
    ClearScreen();

    //  Terminate sound system.
    StopSoundLoop();
    TerminateSound();
    
    //  Terminate libVCT.
    VCT_Cleanup();
#ifdef SDK_PROFILE_FUNCTIONCOST
    OS_DumpStatistics(&myStatBuffer);
#endif    

    BroadCastStatus(DEMO_STATUS_EXIT);
    
    DestroyMenu();
    DWC_SetReportLevel(DWC_REPORTFLAG_ALL);

    return returnSeq;
}

void TransceiverModeDispCallback(int curIdx, void *param)
{
#pragma unused(curIdx, param)
    dbs_CPrintf(NULL, "\n my AID=%d\n", DWC_GetMyAID());
}

/////////////////////////////////////////////////////////////////////////////////////
//
//  Menus
//

static Menu* InitMenu()
{
    if (sMenu == NULL) {
        sMenu = new Menu(NULL);

        sMenu->AddItem(new AudioInfo);
        sMenu->AddItem(new VADInfo);

#ifdef SDK_DEBUG
        //
        //  Add debugging menu on SDK_DEBUG 
        //
        sMenu->AddItem(new AudioDebug);
#endif
    }

    return sMenu;
}

static void DestroyMenu()
{
    delete sMenu;
    sMenu = NULL;
}
