#include <nitro.h>
#include <ninet/ip.h>
#include <dwc.h>

#include "main.h"
#include "init.h"
#include "utility.h"
#include "registerfriend.h"
#include "transport.h"
#include "dbs.h"

#include "screen.h"
#include "userdata.h"


//----------------------------------------------------------------------------
// define
//----------------------------------------------------------------------------

#define GAME_FRAME       1         //  Game frame (1/60s)

static u8 stNumBackup;
static u16 card_lock_id;

static const struct
{
    CARDBackupType type;
    const char *comment;
} stBackupTypeTable[] = {
    { CARD_BACKUP_TYPE_FLASH_2MBITS,    "FLASH     2 Mb" },
    { CARD_BACKUP_TYPE_EEPROM_4KBITS,   "EEPROM    4 kb" },
    { CARD_BACKUP_TYPE_EEPROM_64KBITS,  "EEPROM   64 kb" },
    { CARD_BACKUP_TYPE_EEPROM_512KBITS, "EEPROM  512 kb" },
    { CARD_BACKUP_TYPE_FRAM_256KBITS,   "FRAM    256 kb" },
} ;
enum { BACKUP_TABLE_MAX = sizeof(stBackupTypeTable) / sizeof(*stBackupTypeTable) } ;


//----------------------------------------------------------------------------
// variable
//----------------------------------------------------------------------------
extern DWCUserData stUserData;

extern GameSequence gameSeqList[GAME_MODE_NUM];
extern GameControl stGameCnt;  //  Game control data structure
extern KeyControl stKeyCnt;    //  Key input control structure

/*---------------------------------------------------------------------------*
  V-blank interrupt function
 *---------------------------------------------------------------------------*/
static void GameWaitVBlankIntr(void)
{
    int i;

    dbs_DemoUpdate();

    // Change number of V-Blank wait count with game frame
    for (i = 0; i < GAME_FRAME; i++){
        OS_WaitIrq(TRUE, OS_IE_V_BLANK);
    }

    dbs_DemoLoad();

    // Check for stack overflow
    OS_CheckStack(OS_GetCurrentThread());
}

/*---------------------------------------------------------------------------*
  Display communication error
 *---------------------------------------------------------------------------*/
static void DispErrorMessage0(DWCError error, int errorCode, DWCErrorType errorType)
{

    if (error == DWC_ERROR_NONE) return;

    OS_TPrintf("ERROR: %d, code:%d, type:%d\n", error, -1*errorCode, errorType);
    
    switch (error){
    case DWC_ERROR_DS_MEMORY_ANY:
        OS_TPrintf("ERROR: DS_MEMORY_ANY.\n");
        break;
    case DWC_ERROR_AUTH_ANY:
        OS_TPrintf("ERROR: AUTH_ANY.\n");
        break;
    case DWC_ERROR_AUTH_OUT_OF_SERVICE:
        OS_TPrintf("ERROR: This game's network service has ended.\n");
        break;
    case DWC_ERROR_AUTH_STOP_SERVICE:
        OS_TPrintf("ERROR: Now server is temporarily unavailable.\nPlease re-login later.\n");
        break;
    case DWC_ERROR_AC_ANY:
        OS_TPrintf("ERROR: AC ANY.\n");
        break;
    case DWC_ERROR_NETWORK:
        OS_TPrintf("ERROR: Network error.\n");
        break;
    case DWC_ERROR_DISCONNECTED:
        OS_TPrintf("ERROR: Connection to AP was lost.\n");
        break;
    case DWC_ERROR_FATAL:
        OS_TPrintf("FATAL ERROR: Please turn the power off.\n");
        break;
    case DWC_ERROR_FRIENDS_SHORTAGE:
        OS_TPrintf("ERROR: Your friends seem not to login yet.\n");
        break;
    case DWC_ERROR_NOT_FRIEND_SERVER:
        OS_TPrintf("ERROR: He is not buddy or not game server.\n");
        break;
    case DWC_ERROR_MO_SC_CONNECT_BLOCK:
        OS_TPrintf("ERROR: Game server is busy now.\n");
        break;
    case DWC_ERROR_SERVER_FULL:
        OS_TPrintf("ERROR: Game server is fully occupied.\n");
        break;
    default:
        OS_TPrintf("ERROR: Undefined yet.\n");
        break;
    }
}


/*---------------------------------------------------------------------------*
  Display DWC library error
 *---------------------------------------------------------------------------*/
static DWCError HandleDWCError(GameMode* gameMode)
{
    int errorCode;
    DWCError error;
    DWCErrorType errorType;

    // Get DWC error
    error = DWC_GetLastErrorEx(&errorCode, &errorType);

    // No error, do nothing
    if (error == DWC_ERROR_NONE) return DWC_ERROR_NONE;

    // Display error
    DispErrorMessage0(error, errorCode, errorType);

    DWC_ClearError();        //  Clear error code
    stGameCnt.blocking = 0;  //  Cancel blocking

    switch (errorType){
    case DWC_ETYPE_LIGHT:
        *gameMode = GAME_MODE_LOGIN;
        break;

    case DWC_ETYPE_SHUTDOWN_FM:
        // Terminate matchmaking and free GameSpy heap area
        DWC_ShutdownFriendsMatch();

        // Return to initial state
        *gameMode = GAME_MODE_MAIN;
        break;

    case DWC_ETYPE_DISCONNECT:
        // Terminate matchmaking and free GameSpy heap area
        DWC_ShutdownFriendsMatch();

        // Reset internet connection
        DWC_CleanupInet();

        // Return to initial state
        *gameMode = GAME_MODE_MAIN;
        break;

    case DWC_ETYPE_FATAL:
        // Terminate program with fatal error
        OS_Panic("========== Stop process ==========\n");
        break;
    }

    return error;
}

/*---------------------------------------------------------------------------*
  Backup userdata function
 *---------------------------------------------------------------------------*/
GameMode BackupUserdataMain( void )
{
    GameSequence *gameSeq = &gameSeqList[GAME_MODE_BACKUP];
    int curIdx = 0;
    GameMode returnSeq = GAME_MODE_NUM;
    char userdata[4000];
    u32 userdata_len;

    while (1){
        dbs_Print( 0, 0, "s:%d", DWC_GetMatchingState() );
        dbs_Print( 7, 0, "n:%d", DWC_GetNumConnectionHost() );
        dbs_Print( 0, 1, "w:%d", DWC_GetLinkLevel() );
        dbs_Print( 10,1, "p:%d", stGameCnt.userData.profileID );

        ReadKeyData();  //  Get key data

        HandleDWCError(&returnSeq);  //  Handle error

        // Escape if the next mode to go to has been set
        if (returnSeq != GAME_MODE_NUM) break;

        ////////// Process the following key operations
        if (stKeyCnt.trg & PAD_BUTTON_A){
            // Select the menu with the A button
            BOOL end = FALSE;
            BOOL error = FALSE;
            int last_result;
            CARD_LockBackup((u16)card_lock_id);
            end = CARD_IdentifyBackup(stBackupTypeTable[stNumBackup].type);
            if(!end)
            {
                error = TRUE;
                last_result = CARD_GetResultCode();
            }
            CARD_UnlockBackup(card_lock_id);


            switch (curIdx){
            case 0:  // Load
                if ( error ) {
                    dbs_CPrintf(NULL, "Error occur with %d\n", last_result);
                    continue;
                }

				userdata_len = sizeof(stUserData) + sizeof(stGameCnt.friendList.keyList);
                dbs_CPrintf(NULL, "Backup Load Size = %d\n", userdata_len);
				
                CARD_LockBackup((u16)card_lock_id);
                if (CARD_IsBackupEeprom())
                {
                    CARD_ReadEeprom(0, userdata, userdata_len );
                }
                else if (CARD_IsBackupFlash())
                {
                    CARD_ReadFlash(0, userdata, userdata_len );
                }
                else if (CARD_IsBackupFram())
                {
                    CARD_ReadFram(0, userdata, userdata_len );
                }
                CARD_UnlockBackup(card_lock_id);
                last_result = CARD_GetResultCode();

                if (last_result != CARD_RESULT_SUCCESS)
                {
                    dbs_CPrintf(NULL, "Error occur with %d\n", last_result);
                    continue;
                }
                else
                {

					memcpy(&stUserData, userdata, sizeof(stUserData) );
					memcpy(stGameCnt.friendList.keyList, userdata+sizeof(stUserData), sizeof(stGameCnt.friendList.keyList) );

                    dbs_CPrintf(NULL, "Load OK!\n");
                    returnSeq = GAME_MODE_MAIN;
                }
                break;
            case 1:  // Save
                if ( error ) {
                    dbs_CPrintf(NULL, "Error occur with %d\n", last_result);
                    continue;
                }
                
				userdata_len = sizeof(stUserData) + sizeof(stGameCnt.friendList.keyList);
                dbs_CPrintf(NULL, "Backup Save Size = %d\n", userdata_len);
				memcpy(userdata, &stUserData, sizeof(stUserData) );
				memcpy(userdata+sizeof(stUserData), stGameCnt.friendList.keyList, sizeof(stGameCnt.friendList.keyList) );
                
                CARD_LockBackup((u16)card_lock_id);
                if (CARD_IsBackupEeprom())
                {
                    CARD_WriteAndVerifyEeprom(0, userdata, userdata_len );
                }
                else if (CARD_IsBackupFlash())
                {
                    CARD_WriteAndVerifyFlash(0, userdata, userdata_len );
                }
                else if (CARD_IsBackupFram())
                {
                    CARD_WriteAndVerifyFram(0, userdata, userdata_len );
                }
                CARD_UnlockBackup(card_lock_id);
                last_result = CARD_GetResultCode();

                if (last_result != CARD_RESULT_SUCCESS)
                {
                    dbs_CPrintf(NULL, "Error occur with %d\n", last_result);
                    continue;
                }
                else
                {
                    dbs_CPrintf(NULL, "Save OK!\n");
                    returnSeq = GAME_MODE_MAIN;
                }
                break;
            case 2: // report
                DWC_ReportUserData( &stUserData );
                break;
            case 3:
                returnSeq = GAME_MODE_MAIN;
                break;
            }
        }
        else if (stKeyCnt.trg & PAD_KEY_UP){
            // Move the cursor using up on the crosskey
            curIdx--;
            if (curIdx < 0) curIdx = gameSeq->numMenu-1;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_DOWN){
            // Move the cursor using down on the crosskey
            curIdx++;
            if (curIdx >= gameSeq->numMenu) curIdx = 0;
            // Redisplay menu list
            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_LEFT){
            // Decrement the number of entries using the left button on the crosskey
            stNumBackup--;
            stNumBackup %= BACKUP_TABLE_MAX;

            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        else if (stKeyCnt.trg & PAD_KEY_RIGHT){
            // Increment the number of entries using the right button on the crosskey
            stNumBackup++;
            stNumBackup %= BACKUP_TABLE_MAX;

            DispMenuMsgWithCursor(gameSeq, curIdx, NULL);
        }
        ////////// Key operation processing ends here

        // Wait for V blank interrupt
        GameWaitVBlankIntr();
    }

    return returnSeq;
}


/*---------------------------------------------------------------------------*
  Callback function for making additional displays during backup mode
 *---------------------------------------------------------------------------*/
void BackupModeDispCallback(int curIdx, void* param)
{
#pragma unused(param)
#pragma unused(curIdx)
    dbs_CPrintf(NULL, "\n BACKUP: %s\n", stBackupTypeTable[stNumBackup].comment);
}

